"""Utilities for dynamically adding extra fields to Pydantic models, specifically Norwegian descriptions."""

import logging
from pathlib import Path
from typing import cast

import yaml
from pydantic import BaseModel
from pydantic import Field
from pydantic.config import JsonDict

from dapla_metadata.variable_definitions._utils.config import get_descriptions_path

logger = logging.getLogger(__name__)


def get_package_root() -> Path:
    """Get an absolute Path to the root of the package (dapla_metadata)."""
    number_of_directories_up_from_descriptions_file = 2
    return (
        Path(__file__)
        .resolve()
        .parents[number_of_directories_up_from_descriptions_file]
    )


def load_descriptions(file_path: Path) -> dict:
    """Load and return the contents of a YAML file as a dictionary.

    Args:
        file_path (Path): Path to the YAML file.

    Returns:
    dict: Parsed contents of the YAML file.
    """
    with Path.open(file_path, encoding="utf-8") as f:
        return yaml.safe_load(f)


def apply_norwegian_descriptions_to_model(
    model: type[BaseModel],
) -> None:
    """Add Norwegian descriptions to the fields of a Pydantic model.

    This function globally modifies the model fields by inserting a Norwegian description
    from a predefined dictionary. If a field does not have a corresponding
    Norwegian description, a default message is used.

    Args:
        model (BaseModel): A Pydantic model instance to be updated.

    Returns:
        None: The function modifies the model in place.
    """
    new_fields = {}

    descriptions = load_descriptions(
        get_package_root() / get_descriptions_path(),
    )

    for field_name, field_info in model.model_fields.items():
        new_description: str = descriptions.get(
            field_name,
            f"No description in norwegian found for {field_name}",
        )
        if "No description in norwegian found" in new_description:
            logger.warning("Missing description for %s", field_name)
        else:
            logger.debug("Field %s: %s", field_name, new_description)

        new_fields[field_name] = Field(  # type: ignore[call-overload]
            default=field_info.default,
            alias=field_info.alias,
            title=field_info.title,
            description=field_info.description,
            json_schema_extra=cast(
                JsonDict,
                {
                    "norwegian_description": new_description,
                    "annotation": field_info.annotation,
                },
            ),
        )

    model.model_fields.update(new_fields)  # Apply changes
    model.model_rebuild()



from dapla_metadata._shared.config import get_config_item
from dapla_metadata._shared.config import get_dapla_environment
from dapla_metadata._shared.config import get_dapla_group_context
from dapla_metadata._shared.config import get_oidc_token
from dapla_metadata._shared.enums import DaplaEnvironment
from dapla_metadata.variable_definitions._generated.vardef_client.configuration import (
    Configuration,
)

VARDEF_HOST_TEST = "https://metadata.intern.test.ssb.no"
WORKSPACE_DIR = "WORKSPACE_DIR"
VARDEF_DESCRIPTIONS_FILE_PATH = "VARDEF_DESCRIPTIONS_FILE_PATH"
VARDEF_DEFAULT_DESCRIPTION_PATH = (
    "variable_definitions/resources/vardef_model_descriptions_nb.yaml"
)


def get_descriptions_path() -> str:
    """Get the relative file path from the repo root to the Norwegian descriptions.

    First checks the `VARDEF_DESCRIPTIONS_FILE_PATH` environment variable; if not set, returns a default path.

    Returns:
        str: The file path to the descriptions.
    """
    return (
        get_config_item(VARDEF_DESCRIPTIONS_FILE_PATH)
        or VARDEF_DEFAULT_DESCRIPTION_PATH
    )


def get_workspace_dir() -> str | None:
    """Get the path to work directory from workspace environment variable."""
    return get_config_item(WORKSPACE_DIR)


def get_active_group() -> str:
    """Get the group the user currently represents.

    Returns:
        A string with the active group.
    """
    return str(get_dapla_group_context(raising=True))


def get_vardef_host() -> str:
    """Get the host for the server to make requests to.

    Based on standard Dapla platform environment variables, or variables specific
    to Vardef if these are not available.

    Raises:
        NotImplementedError: When Vardef is not available for the current environment.
    """
    match get_dapla_environment():
        case DaplaEnvironment.PROD:
            msg = "Vardef is not available in prod."
            raise NotImplementedError(msg)
        case DaplaEnvironment.TEST:
            return VARDEF_HOST_TEST
        case DaplaEnvironment.DEV:
            msg = "Vardef is not available in dev."
            raise NotImplementedError(msg)
        case _:
            return get_config_item("VARDEF_HOST") or "http://localhost:8080"


def get_vardef_client_configuration() -> Configuration:
    """Build a config to be supplied to the `ApiClient`."""
    return Configuration(
        host=get_vardef_host(),
        access_token=get_oidc_token(raising=True),
    )



"""Lower level file utilities."""

import logging
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING
from typing import cast

import pytz
from pydantic.config import JsonDict
from ruamel.yaml import YAML
from ruamel.yaml import CommentedMap

from dapla_metadata.variable_definitions._generated.vardef_client.models.complete_response import (
    CompleteResponse,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.variable_status import (
    VariableStatus,
)
from dapla_metadata.variable_definitions._utils import config
from dapla_metadata.variable_definitions._utils.constants import (
    MACHINE_GENERATED_FIELDS,
)
from dapla_metadata.variable_definitions._utils.constants import NORWEGIAN_DESCRIPTIONS
from dapla_metadata.variable_definitions._utils.constants import OPTIONAL_FIELD
from dapla_metadata.variable_definitions._utils.constants import OWNER_FIELD_NAME
from dapla_metadata.variable_definitions._utils.constants import REQUIRED_FIELD
from dapla_metadata.variable_definitions._utils.constants import (
    TEMPLATE_SECTION_HEADER_MACHINE_GENERATED,
)
from dapla_metadata.variable_definitions._utils.constants import (
    TEMPLATE_SECTION_HEADER_OWNER,
)
from dapla_metadata.variable_definitions._utils.constants import (
    TEMPLATE_SECTION_HEADER_STATUS,
)
from dapla_metadata.variable_definitions._utils.constants import (
    VARIABLE_DEFINITIONS_DIR,
)
from dapla_metadata.variable_definitions._utils.constants import (
    VARIABLE_STATUS_FIELD_NAME,
)
from dapla_metadata.variable_definitions._utils.descriptions import (
    apply_norwegian_descriptions_to_model,
)
from dapla_metadata.variable_definitions.exceptions import VardefFileError

if TYPE_CHECKING:
    from pydantic import JsonValue

logger = logging.getLogger(__name__)


def _create_file_name(
    base_name: str,
    time_object: str,
    short_name: str | None = None,
    variable_definition_id: str | None = None,
) -> str:
    """Return file name with dynamic timestamp, and shortname and id if available."""
    return (
        "_".join(
            filter(
                None,
                [
                    base_name,
                    short_name,
                    variable_definition_id,
                    time_object,
                ],
            ),
        )
        + ".yaml"
    )


def _get_current_time() -> str:
    """Return a string format date now for filename."""
    timezone = pytz.timezone("Europe/Oslo")
    current_datetime = datetime.now(timezone).strftime("%Y-%m-%dT%H-%M-%S")
    return str(current_datetime)


def _get_workspace_dir() -> Path:
    """Determine the workspace directory."""
    workspace_dir = config.get_workspace_dir()

    if workspace_dir is None:
        msg = "WORKSPACE_DIR is not set. Check your configuration or provide a custom directory."
        raise VardefFileError(msg)
    workspace_dir_path: Path
    if workspace_dir is not None:
        workspace_dir_path = Path(workspace_dir)
        workspace_dir_path.resolve()

        if not workspace_dir_path.exists():
            msg = f"Directory '{workspace_dir_path}' does not exist."
            raise FileNotFoundError(msg)

        if not workspace_dir_path.is_dir():
            msg = f"'{workspace_dir_path}' is not a directory."
            raise NotADirectoryError(msg)
        logger.debug("'WORKSPACE_DIR' value: %s", workspace_dir)
    return workspace_dir_path


def _get_variable_definitions_dir():
    """Get or create the variable definitions directory inside the workspace."""
    workspace_dir = _get_workspace_dir()
    folder_path = workspace_dir / VARIABLE_DEFINITIONS_DIR
    folder_path.mkdir(parents=True, exist_ok=True)
    return folder_path


def _populate_commented_map(
    field_name: str,
    value: str,
    commented_map: CommentedMap,
    model_instance: CompleteResponse,
) -> None:
    """Add data to a CommentedMap."""
    commented_map[field_name] = value
    field = model_instance.model_fields[field_name]
    description: JsonValue = cast(
        JsonDict,
        field.json_schema_extra,
    )[NORWEGIAN_DESCRIPTIONS]
    if description is not None:
        new_description = (
            (REQUIRED_FIELD if field.is_required() else OPTIONAL_FIELD)
            + "\n"
            + str(description)
        )
        commented_map.yaml_set_comment_before_after_key(
            field_name,
            before=new_description,
        )


def _validate_and_create_directory(custom_directory: Path) -> Path:
    """Ensure that the given path is a valid directory, creating it if necessary.

    Args:
        custom_directory (Path): The target directory path.

    Returns:
        Path: The resolved absolute path of the directory.

    Raises:
        ValueError: If the provided path has a file suffix, indicating a file name instead of a directory.
        NotADirectoryError: If the path exists but is not a directory.
        PermissionError: If there are insufficient permissions to create the directory.
        OSError: If an OS-related error occurs while creating the directory.
    """
    custom_directory = Path(custom_directory).resolve()

    if custom_directory.suffix:
        msg = f"Expected a directory but got a file name: %{custom_directory.name}"
        raise ValueError(msg)

    if custom_directory.exists() and not custom_directory.is_dir():
        msg = f"Path exists but is not a directory: {custom_directory}"
        raise NotADirectoryError(msg)

    try:
        custom_directory.mkdir(parents=True, exist_ok=True)
    except PermissionError as e:
        msg = f"Insufficient permissions to create directory: {custom_directory}"
        raise PermissionError(msg) from e
    except OSError as e:
        msg = f"Failed to create directory {custom_directory}: {e!s}"
        raise OSError(msg) from e

    return custom_directory


def _configure_yaml() -> YAML:
    yaml = YAML()  # Use ruamel.yaml library
    yaml.default_flow_style = False  # Ensures pretty YAML formatting

    yaml.representer.add_representer(
        VariableStatus,
        lambda dumper, data: dumper.represent_scalar(
            "tag:yaml.org,2002:str",
            data.value,
        ),
    )

    return yaml


def _model_to_yaml_with_comments(
    model_instance: CompleteResponse,
    file_name: str,
    start_comment: str,
    custom_directory: Path | None = None,
) -> Path:
    """Convert a model instance to a structured YAML file with Norwegian descriptions as comments.

    Adds Norwegian descriptions to the model, organizes fields into sections, and saves
    the YAML file with a structured format and timestamped filename.

    Args:
        model_instance: The model instance to convert.
        file_name: Name of the generated YAML file.
        start_comment: Comment at the top of the file.
        custom_directory: Optional directory to save the file.

    Returns:
        Path: The file path of the generated YAML file.
    """
    yaml = _configure_yaml()

    from dapla_metadata.variable_definitions.variable_definition import (
        VariableDefinition,
    )

    # Apply new fields to model
    apply_norwegian_descriptions_to_model(VariableDefinition)

    # Convert Pydantic model instance to dictionary
    data = model_instance.model_dump(
        serialize_as_any=True,
        warnings="error",
    )

    # One CommentMap for each section in the yaml file
    machine_generated_map = CommentedMap()
    commented_map = CommentedMap()
    status_map = CommentedMap()
    owner_map = CommentedMap()

    # Loop through all fields in the model and populate the commented maps
    for field_name, value in data.items():
        if field_name == VARIABLE_STATUS_FIELD_NAME:
            _populate_commented_map(field_name, value, status_map, model_instance)
        elif field_name == OWNER_FIELD_NAME:
            _populate_commented_map(field_name, value, owner_map, model_instance)
        elif field_name in MACHINE_GENERATED_FIELDS:
            _populate_commented_map(
                field_name,
                value,
                machine_generated_map,
                model_instance,
            )
        elif field_name not in {VARIABLE_STATUS_FIELD_NAME, OWNER_FIELD_NAME}:
            _populate_commented_map(field_name, value, commented_map, model_instance)

    base_path = (
        _get_variable_definitions_dir()
        if custom_directory is None
        else _validate_and_create_directory(custom_directory)
    )

    file_path = base_path / file_name

    # It is important to preserve the order of the yaml dump operations when writing to file
    # so that the file is predictable for the user
    with file_path.open("w", encoding="utf-8") as file:
        commented_map.yaml_set_start_comment(start_comment)
        yaml.dump(commented_map, file)

        status_map.yaml_set_start_comment(TEMPLATE_SECTION_HEADER_STATUS)
        yaml.dump(status_map, file)

        owner_map.yaml_set_start_comment(TEMPLATE_SECTION_HEADER_OWNER)
        yaml.dump(owner_map, file)

        machine_generated_map.yaml_set_start_comment(
            TEMPLATE_SECTION_HEADER_MACHINE_GENERATED,
        )
        yaml.dump(machine_generated_map, file)
    return file_path
